@@ -18,10 +18,7 @@
  *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
 
-#include <sys/types.h>
-#include <sys/stat.h>
 #include "mdadm.h"
-#include <asm/byteorder.h>
 
 #define min(a,b) (((a) < (b)) ? (a) : (b))
 
@@ -119,24 +116,39 @@ unsigned long long bitmap_bits(unsigned long long array_size,
 
 bitmap_info_t *bitmap_fd_read(int fd, int brief)
 {
+	/* Note: fd might be open O_DIRECT, so we must be
+	 * careful to align reads properly
+	 */
 	unsigned long long total_bits = 0, read_bits = 0, dirty_bits = 0;
 	bitmap_info_t *info;
-	char buf[512];
-	int n;
+	char *buf, *unaligned;
+	int n, skip;
+
+	unaligned = malloc(8192*2);
+	buf = (char*) ((unsigned long)unaligned | 8191)+1;
+	n = read(fd, buf, 8192);
 
 	info = malloc(sizeof(*info));
 	if (info == NULL) {
+#if __GNUC__ < 3
 		fprintf(stderr, Name ": failed to allocate %d bytes\n",
+				(int)sizeof(*info));
+#else
+		fprintf(stderr, Name ": failed to allocate %zd bytes\n",
 				sizeof(*info));
+#endif
 		return NULL;
 	}
 
-	if (read(fd, &info->sb, sizeof(info->sb)) != sizeof(info->sb)) {
+	if (n < sizeof(info->sb)) {
 		fprintf(stderr, Name ": failed to read superblock of bitmap "
 			"file: %s\n", strerror(errno));
 		free(info);
+		free(unaligned);
 		return NULL;
 	}
+	memcpy(&info->sb, buf, sizeof(info->sb));
+	skip = sizeof(info->sb);
 
 	sb_le_to_cpu(&info->sb); /* convert superblock to CPU byte ordering */
 	
@@ -152,23 +164,28 @@ bitmap_info_t *bitmap_fd_read(int fd, int brief)
 	 */
 	total_bits = bitmap_bits(info->sb.sync_size, info->sb.chunksize);
 
-	while ((n = read(fd, buf, sizeof(buf))) > 0) {
+	while(read_bits < total_bits) {
 		unsigned long long remaining = total_bits - read_bits;
 
-		if (remaining > sizeof(buf) * 8) /* we want the full buffer */
-			remaining = sizeof(buf) * 8;
-		if (remaining > n * 8) /* the file is truncated */
-			remaining = n * 8;
-		dirty_bits += count_dirty_bits(buf, remaining);
+		if (n == 0) {
+			n = read(fd, buf, 8192);
+			skip = 0;
+			if (n <= 0)
+				break;
+		}
+		if (remaining > (n-skip) * 8) /* we want the full buffer */
+			remaining = (n-skip) * 8;
+
+		dirty_bits += count_dirty_bits(buf+skip, remaining);
 
 		read_bits += remaining;
-		if (read_bits >= total_bits) /* we've got what we want */
-			break;
+		n = 0;
 	}
 
 	if (read_bits < total_bits) { /* file truncated... */
 		fprintf(stderr, Name ": WARNING: bitmap file is not large "
-			"enough for array size %llu!\n\n", info->sb.sync_size);
+			"enough for array size %llu!\n\n",
+			(unsigned long long)info->sb.sync_size);
 		total_bits = read_bits;
 	}
 out:
@@ -184,14 +201,18 @@ bitmap_info_t *bitmap_file_read(char *filename, int brief, struct supertype **st
 	struct stat stb;
 	struct supertype *st = *stp;
 
-	fd = open(filename, O_RDONLY);
-	if (fd < 0) {
-		fprintf(stderr, Name ": failed to open bitmap file %s: %s\n",
-				filename, strerror(errno));
+	if (stat(filename, &stb) < 0) {
+		fprintf(stderr, Name ": failed to find file %s: %s\n",
+			filename, strerror(errno));
 		return NULL;
 	}
-	fstat(fd, &stb);
 	if ((S_IFMT & stb.st_mode) == S_IFBLK) {
+		fd = open(filename, O_RDONLY);
+		if (fd < 0) {
+			fprintf(stderr, Name ": failed to open bitmap file %s: %s\n",
+				filename, strerror(errno));
+			return NULL;
+		}
 		/* block device, so we are probably after an internal bitmap */
 		if (!st) st = guess_super(fd);
 		if (!st) {
@@ -202,6 +223,13 @@ bitmap_info_t *bitmap_file_read(char *filename, int brief, struct supertype **st
 		}
 		ioctl(fd, BLKFLSBUF, 0); /* make sure we read current data */
 		*stp = st;
+	} else {
+		fd = open(filename, O_RDONLY|O_DIRECT);
+		if (fd < 0) {
+			fprintf(stderr, Name ": failed to open bitmap file %s: %s\n",
+				filename, strerror(errno));
+			return NULL;
+		}
 	}
 
 	info = bitmap_fd_read(fd, brief);
@@ -243,7 +271,8 @@ int ExamineBitmap(char *filename, int brief, struct supertype *st)
 		fprintf(stderr, Name ": invalid bitmap magic 0x%x, the bitmap file appears to be corrupted\n", sb->magic);
 	}
 	printf("         Version : %d\n", sb->version);
-	if (sb->version != BITMAP_MAJOR) {
+	if (sb->version < BITMAP_MAJOR_LO ||
+	    sb->version > BITMAP_MAJOR_HI) {
 		fprintf(stderr, Name ": unknown bitmap version %d, either the bitmap file is corrupted or you need to upgrade your tools\n", sb->version);
 		goto free_info;
 	}
@@ -262,8 +291,8 @@ int ExamineBitmap(char *filename, int brief, struct supertype *st)
 					*(__u32 *)(sb->uuid+8),
 					*(__u32 *)(sb->uuid+12));
 	}
-	printf("          Events : %llu\n", sb->events);
-	printf("  Events Cleared : %llu\n", sb->events_cleared);
+	printf("          Events : %llu\n", (unsigned long long)sb->events);
+	printf("  Events Cleared : %llu\n", (unsigned long long)sb->events_cleared);
 	printf("           State : %s\n", bitmap_state(sb->state));
 	printf("       Chunksize : %s\n", human_chunksize(sb->chunksize));
 	printf("          Daemon : %ds flush period\n", sb->daemon_sleep);
@@ -272,7 +301,7 @@ int ExamineBitmap(char *filename, int brief, struct supertype *st)
 	else
 		sprintf(buf, "Normal");
 	printf("      Write Mode : %s\n", buf);
-	printf("       Sync Size : %llu%s\n", sb->sync_size/2,
+	printf("       Sync Size : %llu%s\n", (unsigned long long)sb->sync_size/2,
 					human_size(sb->sync_size * 512));
 	if (brief)
 		goto free_info;
@@ -287,7 +316,8 @@ free_info:
 int CreateBitmap(char *filename, int force, char uuid[16],
 		 unsigned long chunksize, unsigned long daemon_sleep,
 		 unsigned long write_behind,
-		 unsigned long long array_size /* sectors */)
+		 unsigned long long array_size /* sectors */,
+		 int major)
 {
 	/*
 	 * Create a bitmap file with a superblock and (optionally) a full bitmap
@@ -311,9 +341,20 @@ int CreateBitmap(char *filename, int force, char uuid[16],
 		return rv;
 	}
 
+	if (chunksize == UnSet) {
+		/* We don't want more than 2^21 chunks, as 2^11 fill up one
+		 * 4K page (2 bytes per chunk), and 2^10 address of those
+		 * fill up a 4K indexing page.  2^20 might be safer...
+		 */
+		chunksize = DEFAULT_BITMAP_CHUNK;
+		/* <<21 for 2^21 chunks, >>9 to convert bytes to sectors */
+		while (array_size > (chunksize << (21-9)))
+			chunksize <<= 1;
+	}
+
 	memset(&sb, 0, sizeof(sb));
 	sb.magic = BITMAP_MAGIC;
-	sb.version = BITMAP_MAJOR;
+	sb.version = major;
 	if (uuid != NULL)
 		memcpy(sb.uuid, uuid, 16);
 	sb.chunksize = chunksize;
@@ -349,10 +390,25 @@ int CreateBitmap(char *filename, int force, char uuid[16],
 	
 	rv = 0;
 	/* make the file be the right size (well, to the nearest byte) */
-	ftruncate(fileno(fp), filesize);
+	if (ftruncate(fileno(fp), filesize))
+		perror("ftrunace");
 out:
 	fclose(fp);
 	if (rv)
 		unlink(filename); /* possibly corrupted, better get rid of it */
 	return rv;
 }
+
+void bitmap_update_uuid(int fd, int *uuid)
+{
+	struct bitmap_super_s bm;
+	lseek(fd, 0, 0);
+	if (read(fd, &bm, sizeof(bm)) != sizeof(bm))
+		return;
+	if (bm.magic != __cpu_to_le32(BITMAP_MAGIC))
+		return;
+	memcpy(bm.uuid, uuid, 16);
+	lseek(fd, 0, 0);
+	write(fd, &bm, sizeof(bm));
+	lseek(fd, 0, 0);
+}
