@@ -1,7 +1,7 @@
 /*
  * mdadm - manage Linux "md" devices aka RAID arrays.
  *
- * Copyright (C) 2001-2002 Neil Brown <neilb@cse.unsw.edu.au>
+ * Copyright (C) 2001-2006 Neil Brown <neilb@suse.de>
  *
  *
  *    This program is free software; you can redistribute it and/or modify
@@ -49,7 +49,7 @@ int main(int argc, char *argv[])
 	int level = UnSet;
 	int layout = UnSet;
 	int raiddisks = 0;
-	int max_disks = MD_SB_DISKS;
+	int max_disks = MD_SB_DISKS; /* just a default */
 	int sparedisks = 0;
 	struct mddev_ident_s ident;
 	char *configfile = NULL;
@@ -62,6 +62,7 @@ int main(int argc, char *argv[])
 	int write_behind = 0;
 	int bitmap_fd = -1;
 	char *bitmap_file = NULL;
+	char *backup_file = NULL;
 	int bitmap_chunk = UnSet;
 	int SparcAdjust = 0;
 	mddev_dev_t devlist = NULL;
@@ -74,15 +75,20 @@ int main(int argc, char *argv[])
 	int force = 0;
 	int test = 0;
 	int assume_clean = 0;
-	int autof = 0; /* -2 means create device based on name:
-			*    if it ends mdN, then non-partitioned array N
-			*    if it ends dN, then partitions array N
-			* -1 means create non-partitioned, choose N
-			*  1 or more to create partitioned
-			* If -1 or 1 and name is a 'standard' name, then
-			* insist on a match of type and number.
-			*/
+	/* autof indicates whether and how to create device node.
+	 * bottom 3 bits are style.  Rest (when shifted) are number of parts
+	 * 0  - unset
+	 * 1  - don't create (no)
+	 * 2  - if is_standard, then create (yes)
+	 * 3  - create as 'md' - reject is_standard mdp (md)
+	 * 4  - create as 'mdp' - reject is_standard md (mdp)
+	 * 5  - default to md if not is_standard (md in config file)
+	 * 6  - default to mdp if not is_standard (part, or mdp in config file)
+	 */
+	int autof = 0;
 
+	char *homehost = NULL;
+	char sys_hostname[256];
 	char *mailaddr = NULL;
 	char *program = NULL;
 	int delay = 0;
@@ -94,6 +100,7 @@ int main(int argc, char *argv[])
 	int re_add = 0;
 	char *shortopt = short_options;
 	int dosyslog = 0;
+	int auto_update_home = 0;
 
 	int copies;
 	int print_help = 0;
@@ -111,6 +118,7 @@ int main(int argc, char *argv[])
 	ident.autof = 0;
 	ident.st = NULL;
 	ident.bitmap_fd = -1;
+	ident.bitmap_file = NULL;
 	ident.name[0] = 0;
 
 	while ((option_index = -1) ,
@@ -148,6 +156,10 @@ int main(int argc, char *argv[])
 			}
 			continue;
 
+		case HomeHost:
+			homehost = optarg;
+			continue;
+
 		case ':':
 		case '?':
 			fputs(Usage, stderr);
@@ -166,7 +178,7 @@ int main(int argc, char *argv[])
 		case 'a':
 		case 'r':
 		case 'f':
-		case 6: /* re-add */
+		case ReAdd: /* re-add */
 			if (!mode) {
 				newmode = MANAGE;
 				shortopt = short_bitmap_auto_options;
@@ -226,6 +238,7 @@ int main(int argc, char *argv[])
 					dv->disposition = devmode;
 					dv->writemostly = writemostly;
 					dv->re_add = re_add;
+					dv->used = 0;
 					dv->next = NULL;
 					*devlistend = dv;
 					devlistend = &dv->next;
@@ -303,6 +316,9 @@ int main(int argc, char *argv[])
 			}
 			continue;
 
+		case O(ASSEMBLE,AutoHomeHost):
+			auto_update_home = 1;
+			continue;
 		case O(CREATE,'e'):
 		case O(ASSEMBLE,'e'):
 		case O(MISC,'e'): /* set metadata (superblock) information */
@@ -401,16 +417,18 @@ int main(int argc, char *argv[])
 				break;
 
 			case 10:
-				/* 'f' or 'n' followed by a number <= raid_disks */
-				if ((optarg[0] !=  'n' && optarg[0] != 'f') ||
+				/* 'f', 'o' or 'n' followed by a number <= raid_disks */
+				if ((optarg[0] !=  'n' && optarg[0] != 'f' && optarg[0] != 'o') ||
 				    (copies = strtoul(optarg+1, &cp, 10)) < 1 ||
 				    copies > 200 ||
 				    *cp) {
-					fprintf(stderr, Name ": layout for raid10 must be 'nNN' or 'fNN' where NN is a number, not %s\n", optarg);
+					fprintf(stderr, Name ": layout for raid10 must be 'nNN', 'oNN' or 'fNN' where NN is a number, not %s\n", optarg);
 					exit(2);
 				}
 				if (optarg[0] == 'n')
 					layout = 256 + copies;
+				else if (optarg[0] == 'o')
+					layout = 0x10000 + (copies<<8) + 1;
 				else
 					layout = 1 + (copies<<8);
 				break;
@@ -434,8 +452,8 @@ int main(int argc, char *argv[])
 			}
 			continue;
 
-		case O(CREATE,3):
-		case O(BUILD,3): /* assume clean */
+		case O(CREATE,AssumeClean):
+		case O(BUILD,AssumeClean): /* assume clean */
 			assume_clean = 1;
 			continue;
 
@@ -478,39 +496,7 @@ int main(int argc, char *argv[])
 		case O(CREATE,'a'):
 		case O(BUILD,'a'):
 		case O(ASSEMBLE,'a'): /* auto-creation of device node */
-			if (optarg == NULL)
-				autof = -2;
-			else if (strcasecmp(optarg,"no")==0)
-				autof = 0;
-			else if (strcasecmp(optarg,"yes")==0)
-				autof = -2;
-			else if (strcasecmp(optarg,"md")==0)
-				autof = -1;
-			else {
-				/* There might be digits, and maybe a hypen, at the end */
-				char *e = optarg + strlen(optarg);
-				int num = 4;
-				int len;
-				while (e > optarg && isdigit(e[-1]))
-					e--;
-				if (*e) {
-					num = atoi(e);
-					if (num <= 0) num = 1;
-				}
-				if (e > optarg && e[-1] == '-')
-					e--;
-				len = e - optarg;
-				if ((len == 3 && strncasecmp(optarg,"mdp",3)==0) ||
-				    (len == 1 && strncasecmp(optarg,"p",1)==0) ||
-				    (len >= 4 && strncasecmp(optarg,"part",4)==0))
-					autof = num;
-				else {
-					fprintf(stderr, Name ": --auto flag arg of \"%s\" unrecognised: use no,yes,md,mdp,part\n"
-						"        optionally followed by a number.\n",
-						optarg);
-					exit(2);
-				}
-			}
+			autof = parse_auto(optarg, "--auto flag", 0);
 			continue;
 
 		case O(BUILD,'f'): /* force honouring '-n 1' */
@@ -585,6 +571,10 @@ int main(int argc, char *argv[])
 				continue;
 			if (strcmp(update, "uuid")==0)
 				continue;
+			if (strcmp(update, "name")==0)
+				continue;
+			if (strcmp(update, "homehost")==0)
+				continue;
 			if (strcmp(update, "byteorder")==0) {
 				if (ss) {
 					fprintf(stderr, Name ": must not set metadata type with --update=byteorder.\n");
@@ -602,6 +592,10 @@ int main(int argc, char *argv[])
 			fprintf(stderr, Name ": '--update %s' invalid.  Only 'sparc2.2', 'super-minor', 'uuid', 'resync' or 'summaries' supported\n",update);
 			exit(2);
 
+		case O(ASSEMBLE,NoDegraded): /* --no-degraded */
+			runstop = -1; /* --stop isn't allowed for --assemble, so we overload slightly */
+			continue;
+
 		case O(ASSEMBLE,'c'): /* config file */
 		case O(MISC, 'c'):
 		case O(MONITOR,'c'):
@@ -680,7 +674,7 @@ int main(int argc, char *argv[])
 			devmode = 'a';
 			re_add = 0;
 			continue;
-		case O(MANAGE,6):
+		case O(MANAGE,ReAdd):
 			devmode = 'a';
 			re_add = 1;
 			continue;
@@ -744,7 +738,7 @@ int main(int argc, char *argv[])
 			test = 1;
 			continue;
 
-		case O(MISC, 22):
+		case O(MISC, Sparc22):
 			if (devmode != 'E') {
 				fprintf(stderr, Name ": --sparc2.2 only allowed with --examine\n");
 				exit(2);
@@ -769,6 +763,18 @@ int main(int argc, char *argv[])
 			ident.bitmap_fd = bitmap_fd; /* for Assemble */
 			continue;
 
+		case O(ASSEMBLE, BackupFile):
+		case O(GROW, BackupFile):
+			/* Specify a file into which grow might place a backup,
+			 * or from which assemble might recover a backup
+			 */
+			if (backup_file) {
+				fprintf(stderr, Name ": backup file already specified, rejecting %s\n", optarg);
+				exit(2);
+			}
+			backup_file = optarg;
+			continue;
+
 		case O(GROW,'b'):
 		case O(BUILD,'b'):
 		case O(CREATE,'b'): /* here we create the bitmap */
@@ -782,9 +788,9 @@ int main(int argc, char *argv[])
 			fprintf(stderr, Name ": bitmap file must contain a '/', or be 'internal', or 'none'\n");
 			exit(2);
 
-		case O(GROW,4):
-		case O(BUILD,4):
-		case O(CREATE,4): /* bitmap chunksize */
+		case O(GROW,BitmapChunk):
+		case O(BUILD,BitmapChunk):
+		case O(CREATE,BitmapChunk): /* bitmap chunksize */
 			bitmap_chunk = strtol(optarg, &c, 10);
 			if (!optarg[0] || *c || bitmap_chunk < 0 ||
 					bitmap_chunk & (bitmap_chunk - 1)) {
@@ -796,8 +802,8 @@ int main(int argc, char *argv[])
 			bitmap_chunk = bitmap_chunk ? bitmap_chunk * 1024 : 512;
 			continue;
 
-		case O(BUILD, 5):
-		case O(CREATE, 5): /* write-behind mode */
+		case O(BUILD, WriteBehind):
+		case O(CREATE, WriteBehind): /* write-behind mode */
 			write_behind = DEFAULT_MAX_WRITE_BEHIND;
 			if (optarg) {
 				write_behind = strtol(optarg, &c, 10);
@@ -812,8 +818,13 @@ int main(int argc, char *argv[])
 		/* We have now processed all the valid options. Anything else is
 		 * an error
 		 */
-		fprintf(stderr, Name ": option %c not valid in %s mode\n",
-			opt, map_num(modes, mode));
+		if (option_index > 0)
+			fprintf(stderr, Name ":option --%s not valid in %s mode\n",
+				long_options[option_index].name,
+				map_num(modes, mode));
+		else
+			fprintf(stderr, Name ": option -%c not valid in %s mode\n",
+				opt, map_num(modes, mode));
 		exit(2);
 
 	}
@@ -866,6 +877,8 @@ int main(int argc, char *argv[])
 			fprintf(stderr, Name ": --super-minor=dev is incompatible with --auto\n");	
 			exit(2);
 		}
+		if (mode == MANAGE && runstop < 0)
+			autof=1; /* Don't create */
 		mdfd = open_mddev(devlist->devname, autof);
 		if (mdfd < 0)
 			exit(1);
@@ -897,6 +910,15 @@ int main(int argc, char *argv[])
 		}
 	}
 
+	if (homehost == NULL)
+		homehost = conf_get_homehost(configfile);
+	if (homehost && strcmp(homehost, "<system>")==0) {
+		if (gethostname(sys_hostname, sizeof(sys_hostname)) == 0) {
+			sys_hostname[sizeof(sys_hostname)-1] = 0;
+			homehost = sys_hostname;
+		}
+	}
+
 	rv = 0;
 	switch(mode) {
 	case MANAGE:
@@ -927,20 +949,24 @@ int main(int argc, char *argv[])
 					rv |= 1;
 				else {
 					rv |= Assemble(ss, devlist->devname, mdfd, array_ident, configfile,
-						       NULL,
-						       readonly, runstop, update, verbose-quiet, force);
+						       NULL, backup_file,
+						       readonly, runstop, update, homehost, verbose-quiet, force);
 					close(mdfd);
 				}
 			}
 		} else if (!scan)
 			rv = Assemble(ss, devlist->devname, mdfd, &ident, configfile,
-				      devlist->next,
-				      readonly, runstop, update, verbose-quiet, force);
+				      devlist->next, backup_file,
+				      readonly, runstop, update, homehost, verbose-quiet, force);
 		else if (devs_found>0) {
 			if (update && devs_found > 1) {
 				fprintf(stderr, Name ": can only update a single array at a time\n");
 				exit(1);
 			}
+			if (backup_file && devs_found > 1) {
+				fprintf(stderr, Name ": can only assemble a single array when providing a backup file.\n");
+				exit(1);
+			}
 			for (dv = devlist ; dv ; dv=dv->next) {
 				mddev_ident_t array_ident = conf_get_ident(configfile, dv->devname);
 				if (array_ident == NULL) {
@@ -956,38 +982,103 @@ int main(int argc, char *argv[])
 					continue;
 				}
 				rv |= Assemble(ss, dv->devname, mdfd, array_ident, configfile,
-					       NULL,
-					       readonly, runstop, update, verbose-quiet, force);
+					       NULL, backup_file,
+					       readonly, runstop, update, homehost, verbose-quiet, force);
 				close(mdfd);
 			}
 		} else {
 			mddev_ident_t array_list =  conf_get_ident(configfile, NULL);
-			if (!array_list) {
+			mddev_dev_t devlist = conf_get_devs(configfile);
+			int cnt = 0;
+			if (devlist == NULL) {
+				fprintf(stderr, Name ": No devices listed in conf file were found.\n");
+				exit(1);
+			}
+			if (update) {
+				fprintf(stderr, Name ": --update not meaningful with a --scan assembly.\n");
+				exit(1);
+			}
+			if (backup_file) {
+				fprintf(stderr, Name ": --backup_file not meaningful with a --scan assembly.\n");
+				exit(1);
+			}
+			for (; array_list; array_list = array_list->next) {
+				mdu_array_info_t array;
+				mdfd = open_mddev(array_list->devname,
+						  array_list->autof ? array_list->autof : autof);
+				if (mdfd < 0) {
+					rv |= 1;
+					continue;
+				}
+				if (ioctl(mdfd, GET_ARRAY_INFO, &array)>=0)
+					/* already assembled, skip */
+					;
+				else {
+					rv |= Assemble(ss, array_list->devname, mdfd,
+						       array_list, configfile,
+						       devlist, NULL,
+						       readonly, runstop, NULL, homehost, verbose-quiet, force);
+					if (rv == 0) cnt++;
+				}
+				close(mdfd);
+			}
+			if (homehost) {
+				/* Maybe we can auto-assemble something.
+				 * Repeatedly call Assemble in auto-assmble mode
+				 * until it fails
+				 */
+				int rv2;
+				int acnt;
+				ident.autof = autof;
+				do {
+					acnt = 0;
+					do {
+						rv2 = Assemble(ss, NULL, -1,
+							       &ident, configfile,
+							       devlist, NULL,
+							       readonly, runstop, NULL, homehost, verbose-quiet, force);
+						if (rv2==0) {
+							cnt++;
+							acnt++;
+						}
+						if (rv2 == 1)
+							/* found something so even though assembly failed  we
+							 * want to avoid auto-updates
+							 */
+							auto_update_home = 0;
+					} while (rv2!=2);
+					/* Incase there are stacked devices, we need to go around again */
+					devlist = conf_get_devs(configfile);
+				} while (acnt);
+				if (cnt == 0 && auto_update_home && homehost) {
+					/* Nothing found, maybe we need to bootstrap homehost info */
+					do {
+						acnt = 0;
+						do {
+							rv2 = Assemble(ss, NULL, -1,
+								       &ident, configfile,
+								       devlist, NULL,
+								       readonly, runstop, "homehost", homehost, verbose-quiet, force);
+							if (rv2==0) {
+								cnt++;
+								acnt++;
+							}
+						} while (rv2!=2);
+						/* Incase there are stacked devices, we need to go around again */
+						devlist = conf_get_devs(configfile);
+					} while (acnt);
+				}
+				if (cnt == 0 && rv == 0) {
+					fprintf(stderr, Name ": No arrays found in config file or automatically\n");
+					rv = 1;
+				}
+			} else if (cnt == 0 && rv == 0) {
 				fprintf(stderr, Name ": No arrays found in config file\n");
 				rv = 1;
-			} else
-				for (; array_list; array_list = array_list->next) {
-					mdu_array_info_t array;
-					mdfd = open_mddev(array_list->devname, 
-							  array_list->autof ? array_list->autof : autof);
-					if (mdfd < 0) {
-						rv |= 1;
-						continue;
-					}
-					if (ioctl(mdfd, GET_ARRAY_INFO, &array)>=0)
-						/* already assembled, skip */
-						;
-					else
-						rv |= Assemble(ss, array_list->devname, mdfd,
-							       array_list, configfile,
-							       NULL,
-							       readonly, runstop, NULL, verbose-quiet, force);
-					close(mdfd);
-				}
+			}
 		}
 		break;
 	case BUILD:
-		if (bitmap_chunk == UnSet) bitmap_chunk = DEFAULT_BITMAP_CHUNK;
 		if (delay == 0) delay = DEFAULT_BITMAP_DELAY;
 		if (write_behind && !bitmap_file) {
 			fprintf(stderr, Name ": write-behind mode requires a bitmap.\n");
@@ -1015,7 +1106,7 @@ int main(int argc, char *argv[])
 		}
 
 		rv = Create(ss, devlist->devname, mdfd, chunk, level, layout, size<0 ? 0 : size,
-			    raiddisks, sparedisks, ident.name,
+			    raiddisks, sparedisks, ident.name, homehost,
 			    devs_found-1, devlist->next, runstop, verbose-quiet, force, assume_clean,
 			    bitmap_file, bitmap_chunk, write_behind, delay);
 		break;
@@ -1032,7 +1123,7 @@ int main(int argc, char *argv[])
 				fprintf(stderr, Name ": No devices listed in %s\n", configfile?configfile:DefaultConfFile);
 				exit(1);
 			}
-			rv = Examine(devlist, scan?(verbose>1?0:verbose+1):brief, scan, SparcAdjust, ss);
+			rv = Examine(devlist, scan?(verbose>1?0:verbose+1):brief, scan, SparcAdjust, ss, homehost);
 		} else {
 			if (devlist == NULL) {
 				if (devmode=='D' && scan) {
@@ -1047,7 +1138,7 @@ int main(int argc, char *argv[])
 								e->dev);
 							continue;
 						}
-						rv |= Detail(name, verbose>1?0:verbose+1, test);
+						rv |= Detail(name, verbose>1?0:verbose+1, test, homehost);
 						put_md_name(name);
 					}
 				} else	if (devmode == 'S' && scan) {
@@ -1071,7 +1162,7 @@ int main(int argc, char *argv[])
 									e->dev);
 								continue;
 							}
-							mdfd = open_mddev(name, 0);
+							mdfd = open_mddev(name, 1);
 							if (mdfd >= 0) {
 								if (Manage_runstop(name, mdfd, -1, !last))
 									err = 1;
@@ -1092,15 +1183,15 @@ int main(int argc, char *argv[])
 			for (dv=devlist ; dv; dv=dv->next) {
 				switch(dv->disposition) {
 				case 'D':
-					rv |= Detail(dv->devname, brief?1+verbose:0, test); continue;
+					rv |= Detail(dv->devname, brief?1+verbose:0, test, homehost); continue;
 				case 'K': /* Zero superblock */
-					rv |= Kill(dv->devname, force); continue;
+					rv |= Kill(dv->devname, force, quiet); continue;
 				case 'Q':
 					rv |= Query(dv->devname); continue;
 				case 'X':
 					rv |= ExamineBitmap(dv->devname, brief, ss); continue;
 				}
-				mdfd = open_mddev(dv->devname, 0);
+				mdfd = open_mddev(dv->devname, 1);
 				if (mdfd>=0) {
 					switch(dv->disposition) {
 					case 'R':
@@ -1155,7 +1246,8 @@ int main(int argc, char *argv[])
 		} else if (layout != UnSet)
 			rv = Manage_reconfig(devlist->devname, mdfd, layout);
 		else if (size >= 0 || raiddisks)
-			rv = Manage_resize(devlist->devname, mdfd, size, raiddisks);
+			rv = Grow_reshape(devlist->devname, mdfd, quiet, backup_file,
+					  size, level, layout, chunk, raiddisks);
 		else if (bitmap_file) {
 			if (delay == 0) delay = DEFAULT_BITMAP_DELAY;
 			rv = Grow_addbitmap(devlist->devname, mdfd, bitmap_file,
